#!/usr/bin/env python

from os import unlink, dup2, execl, stat, waitpid, WNOHANG
from popen2 import Popen4
from select import error as selecterror
from signal import signal, SIGCHLD, SIG_DFL
import sys
from copy import deepcopy
from sys import exit
from string import split, strip
from syslog import syslog, LOG_INFO, LOG_ERR, LOG_LOCAL0, openlog
from tempfile import mktemp
from time import time, sleep
from select import select
from base64 import encodestring

from elementtree.ElementTree import Element, XML, tostring, SubElement

from sss.lrs import LRData as Data, LRDataSet as DataSet, LRID as ID
from sss.server import Server, AddEvent
from sss.ssslib import ConnectError

mpdpath='/usr/mpi/bin'
outputspool='/tmp'

class ProcessGroup(Data):
    '''ProcessGroup is a subclassed sss.restriction.Data object that implements mpd process groups'''
    def __init__(self,element, pgid):
        Data.__init__(self,element)
        self.element.tag = "process-group"
        if self.element.find("pgid") == None:
            SubElement(self.element, "pgid")
        self.element.find('pgid').text = pgid
        if self.element.find("state") == None:
            SubElement(self.element, "state")
        self.element.find('state').text = 'running'

        self.mpdpid = None
        for attr in ['outfile','errfile','mpdfile','exitfile']:
            setattr(self,attr,mktemp())
        self.MPIStart()
    
    def MPIStart(self):
        '''MPIStart starts the execution of the process group'''

        self.element.find('state').text = 'running'
        syslog(LOG_INFO,"PGid %s Started"%(self.element.find('pgid').text))
        try:
            AddEvent('process-manager','process_start',self.element.find('pgid').text)
        except:
            syslog(LOG_ERR, "Error sending process_start for pgid %s"%(self.element.find('pgid').text))
        # Dump out xml file
        mpdpg = Element("create-process-group")

        mpdpg.attrib['exit_codes_filename'] = self.exitfile

        for tag in ['submitter', 'totalprocs', 'pgid']:
            if self.element.find(tag) != None:
                mpdpg.attrib[tag] = self.element.find(tag).text

        #convert from lrs to mpd xml...this is kinda messy I will need some help in figuring out the right way to do this.
        pselement = SubElement(mpdpg, "process-spec")
        for element in self.element.find("process-spec"):
            if element.tag == 'arg':
                SubElement(pselement, element.tag , value=element.text, idx=element.attrib['idx'])
            elif element.tag == 'limit':
                SubElement(pselement, element.tag , value=element.text, type=element.attrib['type'], name=element.attrib['name'])
            elif element.tag == 'env':
                SubElement(pselement, element.tag , value=element.text, name=element.attrib['name'])
            else:
                pselement.attrib[element.tag] = element.text

        # add check='yes' to hs
        for node in self.element.findall('host-spec'):
            node.attrib['check'] = 'yes'
            for element in node:
                node.attrib[element.tag] = element.text
            del node.attrib['check']

        open(self.mpdfile,'w').write(tostring(mpdpg))
        # Then fix all values
        
        syslog(LOG_INFO, "running :%s:"%("%s/mpdrun.py -f %s "%(mpdpath,self.mpdfile)))
        pid=fork()
        if pid:
            syslog(LOG_INFO, "Got pid %s for pgid %s"%(pid, self.element.find('pgid').text))
            self.mpdpid=pid
        else:
            null=open('/dev/null','r')
            out=open(self.outfile,'w')
            err=open(self.errfile,'w')
            dup2(null.fileno(),sys.__stdin__.fileno())
            dup2(out.fileno(),sys.__stdout__.fileno())
            dup2(err.fileno(),sys.__stderr__.fileno())
            execl("%s/mpdrun.py"%(mpdpath), 'mpdrun','-f', self.mpdfile)

    def MPIFinish(self):
        '''MPIFinish cleans up after the execution of mpdrun has completed'''
        output = Element("output")
        err = Element("error")
        es = Element("exit-status")

        #backing out the encoding stuff until post release
        #output.text = encodestring(open(self.outfile).read())
        #err.text = encodestring(open(self.errfile).read())
        #self.element.attrib['encoding'] = 'base64'
        
        output.text = open(self.outfile).read()
        err.text = open(self.errfile).read()
        try:
            for element in XML(open(self.exitfile).read()).findall('exit-code'):
                try:
                    tmp = Element("exit-code")
                    for (key, val) in element.attrib.items():
                        SubElement(tmp, key)
                        tmp.find(key).text = val
                    es.append(tmp)
                except:
                    syslog(LOG_ERR, "Exit Status exit parse failure for pgid %s"%(self.element.find('pgid').text))
        except IOError:
            syslog(LOG_ERR, "Failed to read exit status file for pgid %s"%(self.element.find('pgid').text))

        self.element.append(output)
        self.element.append(err)
        self.element.append(es)
        
        for file in ['outfile','errfile','mpdfile','exitfile']:
            try:
                unlink(getattr(self,file))
            except:
                syslog(LOG_ERR, "Failed to unlink %s: %s"%(file,getattr(self,file)))
        syslog(LOG_INFO,"PGid %s Exited"%(self.element.find('pgid').text))
        try:
            AddEvent('process-manager','process_end',self.element.find('pgid').text)
        except:
            syslog(LOG_ERR, "Error sending process_end for pgid %s"%(self.element.find('pgid').text))
        # Send out completion
        self.element.find('state').text = "finished"

        
    def Signal(self, signal, scope):
        pid = fork()
        if pid:
            return pid
        syslog(LOG_INFO,"PGid %s signaled with %s"%(self.element.find('pgid').text,signal))
        if scope == 'single':
            s = '-s'
        else:
            s = '-g'
        execl("%s/mpdsigjob.py"%(mpdpath), "mpdsigjob", signal, '-a', self.element.find('pgid').text, s)

    def MPDKill(self):
        pid = fork()
        if pid:
            return pid
        syslog(LOG_INFO,"PGid %s killed"%(self.element.find('pgid').text))
        execl("%s/mpdkilljob.py"%(mpdpath), "mpdkilljob", '-a', self.element.find('pgid').text)

    def Fetch(self,spec=None):
        data = Data.Display(self,spec)
        if self.element.find('.//state').text == 'running':
            old = signal(SIGCHLD, SIG_DFL)
            mp=Popen4("%s/mpdlistjobs -a %s --sss"%(mpdpath,self.element.find('pgid').text))
            mp.wait()
            out=mp.fromchild.readlines()
            for (host,pid,session) in map(split,map(strip,out)):
                pelement = Element("process")
                SubElement(pelement, "host")
                pelement.find("host").text = host
                SubElement(pelement, "pid")
                pelement.find("pid").text = pid
                SubElement(pelement, "session")
                pelement.find("session").text = session
                data.append(pelement)
            signal(SIGCHLD, old)
        return data

    def Match(self, spec):
        if self.element.tag != spec.tag:
            return False
        for field in [field for field in spec.getchildren() if field.tag not in self.containers]:
            if field.tag == 'process':
                continue
            elif not self.MatchField(self.element.find(field.tag), spec.find(field.tag)):
                syslog( LOG_INFO, "Match function: Matchfield failed for tag: %s"%field.tag)
                return False
        for container in self.containers:
            if spec.find(container):
                if not self.MatchContainer(self.element.find(container), spec.find(container)):
                    return False
        return True



class mpd(Server):
    __implementation__ = 'mpdpm'
    __component__ = 'process-manager'
    __timeout__ = 0.5
    __statefields__ = ['pg']
    __validation__ = 0
    __dispatch__ = {'kill-process-group':'XKillPG',
                    'wait-process-group':'pg.Del',
                    'create-process-group':'AddPG',
                    'signal-process-group':'XSigPG',
                    'get-process-group-info':'XGetPG'}
    
    def __setup__(self):
        self.pg = DataSet("process-groups", "process-group", ProcessGroup, ID(), True)
        self.returns = []
        self.ignore = []
        try:
            stat("%s/mpdrun.py"%(mpdpath))
        except OSError:
            openlog('mpdpm',0,LOG_LOCAL0)
            syslog(LOG_ERR, "Error finding mpd. exiting")
            exit(1)

    def AddPG(self, xml, (peer, port)):
        '''This is needed since the pm syntax isnt quite restriction-like'''
        pg = ProcessGroup(xml, str(self.pg.idalloc.get()))
        self.pg.data.append(pg)
        return pg.Fetch(xml)

    def XGetPG(self, xml, (peer, port)):
        return self.pg.Fetch(xml, (peer, port), None)
    
    def XSigPG(self, xml, (peer,port)):
        return self.pg.Get(xml, (peer,port), self.SigPG)

    def SigPG(self, pg, args):
        self.ignore.append(pg.Signal(args['signal'], args.get('scope', 'single')))

    def XKillPG(self, xml, (peer,port)):
        return self.pg.Get(xml, (peer,port), self.KillPG)

    def KillPG(self,pg, args):
        self.ignore.append(pg.MPDKill())

    def __progress__(self):
        num = 0
        try:
            (pid,stat) = waitpid(-1,WNOHANG)
            if pid != 0:
                self.returns.append((pid,stat))
        except OSError, e:
            if e.errno != 10:
                syslog(LOG_ERR, "Waitpid failed with %s"%(e))
        while self.returns:
            (pid,stat) = self.returns.pop()
            if pid in self.ignore:
                self.ignore.remove(pid)
                continue
            pg = [ x for x in self.pg if x.mpdpid == pid ]
            if len(pg) == 1:
                pg[0].MPIFinish()
                num += 1
            elif len(pg) == 0:
                syslog(LOG_ERR, "Got sigchld for unknown pid %s"%(pid))
        return num

    def SigChildHand(self,sig,frame):
            pass

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from os import fork
    from sys import argv,exit
    try:
        (opts,args)=getopt(argv[1:],'dm:',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nmpdpm.py [-d] [-m <path to mpd> ] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    mpath=filter(lambda x:(x[0] == '-m'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if mpath:
        mpdpath=mpath[0][1]
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    s=mpd(debug=ldebug)
    s.ServeForever()
